#!/usr/bin/env python
# -*- coding: utf-8 -*-



import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gimpshelf import shelf
import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################
# -------------------------------------------------------------------------------------
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: Filters/Custom Font Tools/Custom Font Paths to Layers vX...
# It uses named character shaped paths to stroke and create Gimp custom fonts
# The name format consists of the required character plus a space plus the four letters char eg in quotes 'A char', 'B char' etc
# If a character path is followed by another path named 'A part', 'B part' etc that path will be stroked with a differently sized line or brush
# Character layers are generated with the usual name convention using the '.png' suffix
# This can be useful when creating certain punctuation characters such as a comma
# The plug-in will take the layer height from the parent image and attempt to size the character layer width according to the current brush or line size in the filter dialog
# Padding space can be added to the left and right margins of the characters
# The brush option provides the opportunity to create a wide range of effects - although the dialog has an option to alter brush spacing, more variation to stroking can be added 
# by the use of dynamics and animated brushes
# The line option provides a means of estimating the most effective size of brush to work with 
# Existing character layers can be re-stroked with different brushes to create a layered effect
# A space character will be created automatically with the width chosen in the dialog as 'space.png'

# Current version number: 0.07 Tabbed

# Version 0.02 Improved handling of characters created from two paths; catches mis-placed part paths in path stack
# Version 0.03 Further improved handling of characters created from two paths; Added options to remove paths from the file ready for saving as a custom font file
# Version 0.04 added option to allow brush to stroke paths and emulate 'Track Direction' dynamics using code by Tim and a function by Ofnuts to calculate the slope of a point along a path
# Version 0.05 added option to include size jitter and random angle for emulating dynamics
#              added options for lining up paths (not really happy with the convoluted solution) with different sized widths when stroking
# Version 0.06 Combined options for separate line/brush sizes into one stroke size option
#			   Added options for tapering stroke size along path
#              Added replacement of old style two part paths with a single path when using dynamics emulation
# Version 0.07 Added drop shadow options
# Version 0.07 Tabbed - Uses MrQ's port of  Arakne's Gimpui3 for tabbed dialogs in order to save screen space
# Version 0.08 Tabbed - drop shadow can be applied to first layer created
#                     - added preview option to create a single character as a new image
# Version 0.09 Tabbed - bug fix for bruch tracking size



# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)



def do_stuff(image, layer, path_style, brush_width_major, brush_width_minor, brush_spacing, thinner_chars, cap_type, join_type, left_margin, right_margin, space_width, 
            repeat_process, force_dynamics, track_direction, random_angle, size_jitter, space_jitter, track_size, min_track_size, track_size_all, exclude_chars, 
            drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, remember_width, do_remove_all_letter_layers, remove_paths, do_preview, preview_char, ptab):
    
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)
	

	
    add_drop = False
    if drop_offset_x<>0 or drop_offset_y<>0: add_drop = True
	
    line_width_major = brush_width_major
	
    if do_remove_all_letter_layers:
	
    
        loop_ptr = 1
        num_layers, layer_ids = pdb.gimp_image_get_layers(image)
        loop = num_layers - loop_ptr
        layer_names_list = image.layers
	
        visi_loop, other_layers, first_letter_layer, has_a_space = 0, 0, -1, False
        
        while visi_loop < num_layers:
            # check for layers to create new faces for
            visi_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            visi_name = visi_name.decode('utf-8')
            test = fn_test_for_custom_layer(image, layer_names_list[visi_loop], visi_name)
            if test or visi_name == "space.png":
			    # Re-position letter layers
                layer_names_list[visi_loop].set_offsets(0, 0)
                
			    # Remove character layers inc space character layer
                pdb.gimp_image_remove_layer(image, layer_names_list[visi_loop])
                pdb.gimp_image_resize_to_layers(image)
                
                    		
            visi_loop += 1 
	
	
        if shelf.has_key('Paths_to_Layers_v009') <> 0: # if this is not the first time we have stored margin info
            # using Gimpshelf module to store information about layer margins in persistent memory
            # set stroke method to paint line as a line or with the currently selected brush
            if path_style:
                stroke_method = 0
                global_width = line_width_major
            else:
                stroke_method = 1
                global_width = brush_width_major
				
            global_left, global_right = left_margin, right_margin
            if remember_width: 
                #shelf['Paths_to_Layers_v009'] = [str(global_left), str(global_width), str(global_right)]
                shelf['Paths_to_Layers_v009'] = [str(left_margin), str(global_width), str(right_margin)]
                shelf['Paths_to_Layers_v009'] = ["0","0","0"]
            else:
                shelf['Paths_to_Layers_v009'] = ["0","0","0"]
	
    else:
        line_width_major, line_width_minor = brush_width_major, brush_width_minor
        # set stroke method to paint line as a line or with the currently selected brush
        if path_style:
            stroke_method = 0
        else:
            stroke_method = 1
		
        pdb.gimp_context_set_stroke_method(stroke_method) # stroke with brush or line
	
        global_width = 0
	
        if remember_width:
            dummy = 0
            if shelf.has_key('Paths_to_Layers_v009') == 0: # if this is first time
                dummy = 0
                if stroke_method == 0:
                    global_width = line_width_major
                else:
                    global_width = brush_width_major
                global_left, global_right = left_margin, right_margin
                shelf['Paths_to_Layers_v009'] = [str(global_left), str(global_width), str(global_right)]
                all_shelf = shelf['Paths_to_Layers_v009']
                #pdb.gimp_message("First time = " + str(all_shelf) )
            else:
                # if this is not the first time retrieve settings from persistent memory
                if shelf['Paths_to_Layers_v009'] <> 0:
                    all_shelf = shelf['Paths_to_Layers_v009']
                    if all_shelf[1] == "0":
                        dummy = 1
                        if stroke_method == 0:
                            global_width = line_width_major
                        else:
                            global_width = brush_width_major
                        global_left, global_right = left_margin, right_margin
                        shelf['Paths_to_Layers_v009'] = [str(global_left), str(global_width), str(global_right)]
                    else:
                        dummy = 1
                        all_shelf = shelf['Paths_to_Layers_v009']
                        global_width = float(all_shelf[1])
                        global_width = int(global_width)
                        global_left, global_right = float(all_shelf[0]), float(all_shelf[2])
                        global_left, global_right = int(global_left), int(global_right)
        else:
            dummy = 3
            #all_shelf = shelf['Paths_to_Layers_v009']
            if shelf.has_key('Paths_to_Layers_v009') == 0:
                dummy = 5
                #pdb.gimp_message("No Gimpshelf" )
            else:
                shelf['Paths_to_Layers_v009'] = ["0","0","0"]
                ###############
                # set stroke method to paint line as a line or with the currently selected brush
                if path_style:
                    stroke_method = 0
                    global_width = line_width_major
                else:
                    stroke_method = 1
                    global_width = brush_width_major
				
                global_left, global_right = left_margin, right_margin
                shelf['Paths_to_Layers_v009'] = [str(global_left), str(global_width), str(global_right)]
				
				
                all_shelf = shelf['Paths_to_Layers_v009']
                #pdb.gimp_message("Retrieved reset - disremembered = " + str(all_shelf) )
                ###############
		
        if remember_width:
            if (stroke_method == 0 and global_width>line_width_major) or (stroke_method == 1 and global_width>brush_width_major):
                if stroke_method == 0:
                    left_margin = ((global_width/2) + left_margin) - (line_width_major/2)
                    right_margin = global_width - (left_margin + line_width_major)
                else:
                    left_margin = ((global_width/2) + left_margin) - (brush_width_major/2)
                    right_margin = global_width - (left_margin + brush_width_major)
			
    
        #pdb.gimp_message("global_width = " + str(global_width) + " : dummy = " + str(dummy) )
		
        if do_preview:
            proc_stroke_preview(image, line_width_major, line_width_minor, brush_width_major, brush_width_minor, join_type, cap_type, brush_spacing,
                    force_dynamics, thinner_chars, stroke_method, left_margin, right_margin, repeat_process, track_direction,  track_size_all,
                    random_angle, size_jitter, space_jitter, exclude_chars, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, add_drop,
                    track_size, min_track_size, remove_paths, space_width, do_preview, preview_char)
					
        else:
            proc_do_stroking(image, line_width_major, line_width_minor, brush_width_major, brush_width_minor, join_type, cap_type, brush_spacing,
                        force_dynamics, thinner_chars, stroke_method, left_margin, right_margin, repeat_process, track_direction,  track_size_all,
                        random_angle, size_jitter, space_jitter, exclude_chars, 
                        drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, add_drop, track_size, min_track_size, remove_paths, space_width, do_preview)
			
		
        '''
        if do_preview: 
            if pdb.gimp_image_get_layer_by_name(image, "Preview from Visible"):
                pdb.gimp_message("Preview layer exists")
            else:
                pdb.gimp_message("No preview layer exists")
                view = pdb.gimp_layer_new_from_visible(image, image, "Preview from Visible")
                pdb.gimp_image_insert_layer(image, view, None, 0)
        '''
    pdb.gimp_displays_flush()     
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	
	
######################################################################################################################################

def proc_stroke_preview(image, line_width_major, line_width_minor, brush_width_major, brush_width_minor, join_type, cap_type, brush_spacing,
                    force_dynamics, thinner_chars, stroke_method, left_margin, right_margin, repeat_process, track_direction,  track_size_all,
                    random_angle, size_jitter, space_jitter, exclude_chars, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, add_drop,
                    track_size, min_track_size, remove_paths, space_width, do_preview, preview_char): 
					
    if preview_char == "": preview_char = "A"
    preview_char = preview_char + " char"
    num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
    vector_list = image.vectors
	
    # check that the preview character is present or use another
    if preview_char in str(vector_list):
        prev_pos = str(vector_list).find(preview_char)
        prev_loop = str(vector_list).count(",",0,prev_pos)
    else:
        preview_char = pdb.gimp_item_get_name(vector_list[0]) + " char"
        prev_loop = 0
    
	
    # Set various line stroke options
    pdb.gimp_context_set_line_width(line_width_major)
    pdb.gimp_context_set_line_join_style(join_type)
    pdb.gimp_context_set_line_cap_style(cap_type)

    thinner_line_size = int((line_width_major/100) * line_width_minor)
    thinner_brush_size = int((brush_width_major/100) * brush_width_minor)
	
	# set colour for stroking 
    stroke_colour = pdb.gimp_context_get_foreground()
    pdb.gimp_context_set_foreground(stroke_colour)
	
    # set brush options
    pdb.gimp_context_set_sample_transparent(True)
    current_brush = pdb.gimp_context_get_brush()
    pdb.gimp_context_set_brush_size(brush_width_major)
    brush_spacing = float(brush_spacing/100)
    if brush_spacing>0: pdb.gimp_context_set_brush_spacing(brush_spacing)
	
	
    new_vector = vector_list[0]
    vector_copy = pdb.gimp_vectors_copy(new_vector)
	
	
    # add new image
    new_image = pdb.gimp_image_new(image.width, image.height, 0)
	
    # Set start undo point
    pdb.gimp_image_undo_group_start(new_image)
	
    # get vector info for preview character
    preview_char_ListOfControlPoints = []
    stroke_loop = 0
	
    # get preview vector id
    preview_id = pdb.gimp_image_get_vectors_by_name(image, preview_char)
   	
    # export preview vector in string format
    stroke_string = pdb.gimp_vectors_export_to_string(image, preview_id)
		
    preview_char_id = pdb.gimp_vectors_new(new_image, preview_char)
	
    # import the preview vector to the new image
    num_vectors, vectors_ids = pdb.gimp_vectors_import_from_string(new_image, stroke_string, -1, True, False)
	
	
    # get lists of vectors and layers ids in this image
    num_vectors, vector_ids = pdb.gimp_image_get_vectors(new_image)
    vector_list = new_image.vectors
			
    original_layer = pdb.gimp_layer_new(new_image, image.width, image.height, 1, "Preview image background", 100, 28)
    new_image.add_layer(original_layer, -1)
    
    new_display = pdb.gimp_display_new(new_image)	

    layer_list = new_image.layers
    
    # create a list of the actual layer names
    loop = 0
    while loop<len(layer_list):
        pdb.gimp_item_set_visible(layer_list[loop], False)
        layer_name = pdb.gimp_item_get_name(layer_list[loop])
        layer_name = layer_name.decode('utf-8')
        if loop == 0:
            layer_names_list = [layer_name]
        else:
            layer_names_list.append(layer_name)
        loop += 1
    found_valid_vector = False
    found_vector = False
    if num_vectors>0: # if there are any vectors in this new_image do something
        found_vector = True
        loop = 0
        update_canvas_size = False
        previous_name = "undefined"
        while loop<(1): # preview only uses one character
            vector_id = vector_list[loop]
            active_vector = pdb.gimp_image_set_active_vectors(new_image,vector_id)
            active_vector_name = pdb.gimp_item_get_name(vector_id)
            active_vector_name = active_vector_name.decode('utf-8')
				
            ###############
			
            if (len(active_vector_name) == 6 and " char" in active_vector_name) or ((len(active_vector_name) == 6 and " part" in active_vector_name) and previous_name == active_vector_name[0]): # check for compatible paths
			
			
                first_char = active_vector_name[0]
                #pdb.gimp_message("first_char = "+str(first_char))
				
                if first_char in thinner_chars:
                    pdb.gimp_context_set_line_width(thinner_line_size)
                    pdb.gimp_context_set_brush_size(thinner_brush_size)
				
                else:
                    pdb.gimp_context_set_line_width(line_width_major)
                    pdb.gimp_context_set_brush_size(brush_width_major)
					
                found_valid_vector = True
				
                if loop<(num_vectors-1):
                    next_vector_id = vector_list[loop + 1]
                    next_vector_name = pdb.gimp_item_get_name(next_vector_id)
                    next_vector_name = next_vector_name.decode('utf-8')
                else:
                    next_vector_id = -1
                    next_vector_name = "undefined"
				
                if " char" in active_vector_name:
                    new_name = active_vector_name[0]
                    previous_name = new_name
                    new_name = new_name.decode('utf-8')
                    new_name += ".png"
					
                    
                    layer_exists = False
                    if new_name in layer_names_list: #check to see if there is already a layer of this name
                        # and get its layer id
                        existing_layer = pdb.gimp_image_get_layer_by_name(new_image, new_name)
                        layer_exists = True
                    
                    # use image height for layer and get information about the width from points along path - hopefully they will help define the width fairly accurately
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(vector_id)
                    maxX,minX,stroke_loop = -3000000,3000000,0
                    while stroke_loop<NumberOfStrokes_in_Path:
                        Char_Vector_Stroke = Char_Vector_StrokeIDs_List[stroke_loop]
                        TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(vector_id,Char_Vector_Stroke)
                        ploop,cloop = 0,0
                        #make a copy of the returned list of control points
                        Temp_ListOfControlPoints = list(ListOfControlPoints)
                        # remove unnecessary control handle entries retaining x coordinate positions for points
                        while ploop<len(ListOfControlPoints):
                            if ploop == 2 or ((ploop-2)%6 == 0):
                                cloop -= 0
                            else:
                                list_pos = ploop - cloop
                                del Temp_ListOfControlPoints[list_pos]
                                cloop += 1
                            ploop+=1
							
                        # get maximum x position from this list to find layer width
                        if max(Temp_ListOfControlPoints) > maxX: maxX = max(Temp_ListOfControlPoints)
                        # get minimum x position to use to help set layer x offset
                        if min(Temp_ListOfControlPoints) < minX: minX = min(Temp_ListOfControlPoints)
                        stroke_loop += 1
						
                    # use line width or brush width to get a path width according to type of current stroke method
                    if stroke_method == 1:
                        stroke_width = int(brush_width_major/2)
                    else:
                        stroke_width = int(line_width_major/2)
							
                    # set the width of the layer                        
                    layer_left_margin = minX - ((stroke_width) + left_margin) # define x pos of left edge of layer
                    layer_right_margin = maxX + ((stroke_width) + right_margin) # define x pos of right edge of layer
                    y_offset, char_x_offset = 0, 0

                    if layer_left_margin<0:
                        layer_width = abs(layer_left_margin - layer_right_margin)
                        char_x_offset = (layer_left_margin)
                    else:
                        layer_width = layer_right_margin - layer_left_margin
                        char_x_offset = (layer_left_margin)
							
							
                    new_layer = pdb.gimp_layer_new(new_image, layer_width, new_image.height, 1, new_name,100,0)
                    pdb.gimp_image_insert_layer(new_image, new_layer, None, 0)
                    if layer_width>new_layer.width:
                        pdb.gimp_layer_resize(new_layer, layer_width, new_image.height, 0, 0)
                        if layer_width>new_image.width: update_canvas_size = True
							
						
                    # move layer to stroke position
                    pdb.gimp_item_transform_translate(new_layer, char_x_offset, y_offset)


                elif " part" in active_vector_name:
                    # check to see if this is a part layer to stroke in the secondary size
                    new_name = active_vector_name[0] + ".png"
                    part_name = active_vector_name[0]
                    new_name = new_name.decode('utf-8')
                    if not layer_exists: new_layer = pdb.gimp_image_get_layer_by_name(new_image, new_name)
                    if " part" in active_vector_name:
                        # if the path is one of ours set brush/line size
                        pdb.gimp_context_set_line_width(line_width_minor)
                        pdb.gimp_context_set_brush_size(brush_width_minor)
						
                        # use image height for layer and get information about the width from points along path - hopefully they will help define the width fairly accurately
                        part_maxX,part_minX,stroke_loop = -3000000,3000000,0
                        NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(vector_id)
                        while stroke_loop<NumberOfStrokes_in_Path:
                            Char_Vector_Stroke = Char_Vector_StrokeIDs_List[stroke_loop]
                            TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(vector_id,Char_Vector_Stroke)
                            ploop,cloop = 0,0
                            #make a copy of the returned list of control points
                            Temp_ListOfControlPoints = list(ListOfControlPoints)
                            # remove unnecessary control handle entries and retain x coordinate positions for points
                            while ploop<len(ListOfControlPoints):
                                if ploop == 2 or ((ploop-2)%6 == 0):
                                    cloop -= 0
                                else:
                                    list_pos = ploop - cloop
                                    del Temp_ListOfControlPoints[list_pos]
                                    cloop += 1
                                ploop+=1
                            # get maximum x position to use to help set layer width
                            if max(Temp_ListOfControlPoints) > part_maxX: part_maxX = max(Temp_ListOfControlPoints)
                            # get minimum x position to use to help set layer x offset
                            if min(Temp_ListOfControlPoints) < part_minX: part_minX = min(Temp_ListOfControlPoints)
                            stroke_loop += 1
							
                        # use line width or brush width to stroke a path
                        if stroke_method == 1:
                            stroke_width = int(brush_width_minor/2)
                        else:
                            stroke_width = int(line_width_minor/2)

                        # check if the part layer will need a wider layer than the parent character layer
                        # set the margin size for the part layers
                        part_layer_right_margin = part_maxX + ((stroke_width) + right_margin) # define x pos of right edge of part layer
                        part_layer_left_margin = part_minX - ((stroke_width) + left_margin) # define x pos of left edge of part layer
                        part_width = layer_width # provisional width
                        
                        if min(layer_left_margin, part_layer_left_margin)<0:
                            part_width = abs(min(layer_left_margin, part_layer_left_margin) - max(layer_right_margin,part_layer_right_margin) )
                        else:
                            part_width = max(layer_right_margin,part_layer_right_margin) - min(layer_left_margin, part_layer_left_margin)
                        # set value for this part layer offset
                        part_x_offset = 0
                        if part_layer_left_margin<layer_left_margin:
                            part_x_offset = abs(part_layer_left_margin - layer_left_margin)
                        else:
                            part_x_offset = 0
							
                        if part_width>layer_width:
                            # resize the character layer if extra width is required
                            pdb.gimp_layer_resize(new_layer, part_width, new_image.height,part_x_offset, y_offset)
                        pdb.gimp_item_transform_translate(new_layer, char_x_offset , y_offset) # move the layer into position for stroking
						
                        if part_width>new_image.width: update_canvas_size = True # set flag for updating canvas size
											
                # make layer visible
                pdb.gimp_item_set_visible(new_layer, True)
		
				
					
                # stroke the path (using current width/colour settings) the required number of times
                repeat_loop = 0
				
				
                #if force_dynamics and stroke_method == 1 and " char" in active_vector_name: # move current vector to accomodate changes in left hand margins for track direction option
                if force_dynamics and stroke_method == 1: # move current vector to accomodate changes in left hand margins for track direction option

                    num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(vector_id)
                
                    for i in range(0,num_strokes):
                        pdb.gimp_vectors_stroke_translate(vector_id, stroke_ids[i], (0-layer_left_margin), 0)
                    
                    '''
                    if force_dynamics and stroke_method == 1 and " part" in active_vector_name: # move current vector to accomodate changes in left hand margins for track direction option
                    #if force_dynamics and stroke_method == 1: # move current vector to accomodate changes in left hand margins for track direction option

                        num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(vector_id)
                
                        for i in range(0,num_strokes):
                            #pdb.gimp_vectors_stroke_translate(vector_id, stroke_ids[i], (0-layer_left_margin), 0)
                            pdb.gimp_vectors_stroke_translate(vector_id, stroke_ids[i], (0+(int(brush_width_minor)/2)), 0)
                    '''
                
                
                while repeat_loop<repeat_process:
                    #pdb.gimp_image_undo_group_start(image)
				
				
                    if (force_dynamics and stroke_method == 1):
					
                        # Using a version of the code that Tim made available for using the paintbrush tool to follow direction of a stroke which in turn uses a function written by Ofnuts
                        #pdb.gimp_context_set_dynamics("Track Direction") # Doesn't seem to take effect at all when we try to stroke vectors
                        
                        num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(vector_id)
                        paint_between_length = brush_spacing*50
                        original_brush_size = pdb.gimp_context_get_brush_size()
                        target_brush_size = float(original_brush_size-((original_brush_size/100)*min_track_size ))
                        target_brush_size = float(((original_brush_size/100)*min_track_size ))
                        for i in range(0,num_strokes):
                            step_brush_size = original_brush_size
                            stroke_length = pdb.gimp_vectors_stroke_get_length(vector_id,stroke_ids[i],0.01)
                            paint_between_length = float(brush_spacing*50)
                            num_brush_steps = float(stroke_length/paint_between_length)
                            if stroke_length % brush_spacing == 0: paint_between_length -= 1
                            dist_at = 0.0
                            if stroke_length > 0.0:size_step = float((original_brush_size-target_brush_size)/num_brush_steps)
                            
                            while dist_at < (stroke_length-1):
                                x1,y1,slope,valid = pdb.gimp_vectors_stroke_get_point_at_dist(vector_id,stroke_ids[i],dist_at-1,0.01)
                                x2,y2,slope,valid = pdb.gimp_vectors_stroke_get_point_at_dist(vector_id,stroke_ids[i],dist_at,0.01)
                                x3,y3,slope,valid = pdb.gimp_vectors_stroke_get_point_at_dist(vector_id,stroke_ids[i],dist_at+1,0.01)
         
                                points = []
                                points.append(x2)
                                points.append(y2)
                                dist_at += paint_between_length
								
                                if track_direction:
                                    angle = computeOrientedSlope(x3-x1,y3-y1,slope) #calculate angle based on slope using function by Ofnuts
                                    angle = angle/2/math.pi * 360
                                    if angle > 180:
                                        angle -= 360
                                    pdb.gimp_context_set_brush_angle(angle)
									
								
                                if random_angle:
                                    random.seed()
                                    angle = (random.randrange(0, 360)-180.0)
                                    pdb.gimp_context_set_brush_angle(angle)
								
									
                                if size_jitter>0:
                                    if stroke_width>10:
                                        jitter_stroke_width = stroke_width * 2
                                        jitter_adj = int((float(jitter_stroke_width)/100)* size_jitter) 
                                        jitter_size_min, jitter_size_max = jitter_stroke_width - jitter_adj, jitter_stroke_width + jitter_adj
                                        random.seed()
                                        jitter_brush_size = random.randrange(jitter_size_min, jitter_size_max)
                                        if jitter_brush_size<10:jitter_brush_size = 10
                                        pdb.gimp_context_set_brush_size(jitter_brush_size)
										
										
                                if space_jitter>0:
                                    jitter_space_adj = ((brush_spacing*50)/100) * space_jitter
                                    jitter_space_min, jitter_space_max = (brush_spacing*50) - jitter_space_adj, (brush_spacing*50) + jitter_space_adj
                                    random.seed()
                                    jitter_brush_spacing = random.randrange(int(jitter_space_min), int(jitter_space_max))
                                    if jitter_brush_spacing<1:jitter_brush_spacing = 1
                                    paint_between_length = jitter_brush_spacing
									
                                if first_char not in exclude_chars: # is the character in the exclusion string
                                    if ((track_size and i>0) or (track_size and (i==0 and track_size_all))): # check to see for taper option
                                        step_brush_size -= size_step
                                        if step_brush_size<1.00: step_brush_size = 1.0
                                        pdb.gimp_context_set_brush_size(step_brush_size)

									
                                pdb.gimp_paintbrush(new_layer,0,len(points),points,PAINT_CONSTANT,0) # use paintbrush tool to paint along current stroke to emulate track direction
							
					
                    else:
                        pdb.gimp_drawable_edit_stroke_item(new_layer, vector_id) # stroke path when no track direction is used
					
                    #pdb.gimp_image_undo_group_end(image)
                    pdb.gimp_displays_flush()
                    repeat_loop += 1
					
					
                # reset the layer position
                new_layer.set_offsets(0, 0)
				
				
                if "part" not in next_vector_name:
                    pdb.gimp_item_set_visible(new_layer, True)
                    if layer_exists:
                        pdb.gimp_item_set_visible(existing_layer, True)
						
                    if add_drop:
                        drop_layer = new_layer
                        new_layer = proc_dropshadow(new_image, drop_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour)
						
                    if layer_exists: pdb.gimp_image_merge_down(new_image, new_layer, 0)
						
							
						
				
                # turn off layer visibility
                #if loop <> (num_vectors-1): pdb.gimp_item_set_visible(new_layer, False) # turn off layer visibility
                pdb.gimp_displays_flush()
                '''		
                '''
				
            loop += 1
	
    # Set end undo point
    pdb.gimp_image_undo_group_end(new_image)
    
	
######################################################################################################################################


def proc_do_stroking(image, line_width_major, line_width_minor, brush_width_major, brush_width_minor, join_type, cap_type, brush_spacing,
                    force_dynamics, thinner_chars, stroke_method, left_margin, right_margin, repeat_process, track_direction,  track_size_all,
                    random_angle, size_jitter, space_jitter, exclude_chars, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour, add_drop,
                    track_size, min_track_size, remove_paths, space_width, do_preview): 
					
    #### Start of code for proc_do_stroking
	
	
        # Set various line stroke options
    pdb.gimp_context_set_line_width(line_width_major)
    pdb.gimp_context_set_line_join_style(join_type)
    pdb.gimp_context_set_line_cap_style(cap_type)

    thinner_line_size = int((line_width_major/100) * line_width_minor)
    thinner_brush_size = int((brush_width_major/100) * brush_width_minor)
	
	# set colour for stroking 
    stroke_colour = pdb.gimp_context_get_foreground()
    pdb.gimp_context_set_foreground(stroke_colour)
	
    # set brush options
    pdb.gimp_context_set_sample_transparent(True)
    current_brush = pdb.gimp_context_get_brush()
    pdb.gimp_context_set_brush_size(brush_width_major)
    brush_spacing = float(brush_spacing/100)
    if brush_spacing>0: pdb.gimp_context_set_brush_spacing(brush_spacing)
    # get lists of vectors and layers ids in this image
    num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
    vector_list = image.vectors
		
    if " part" in str(vector_list) and force_dynamics: 
		
        ###########
        
        # Since there is no pdb to merge two paths together we need to 
        # deal with character path files that use two character parts i.e. one ending in char that acts as the base character 
        # and one ending in part by joining the part vector onto the char vector
        # get lists of vector ids in image
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
		
        # create a list of the actual vector names
        loop, part_flag = 0, False
		
        while loop<len(vector_list):
            pdb.gimp_item_set_visible(vector_list[loop], False)
            vector_name = pdb.gimp_item_get_name(vector_list[loop])
            vector_name = vector_name.decode('utf-8') # in case of unicode characters 
            if "part" in vector_name: part_flag = True
            if loop == 0:
                vector_names_list = [vector_name]
            else:
                vector_names_list.append(vector_name)
				
            loop += 1
			
        loop = 0
        while loop<len(vector_list):
            vector_name = vector_names_list[loop]
            vector_id = vector_list[loop]
            if (len(vector_name) == 6 and " char" in vector_name): #check that the vector is one we can use
                next_vector_name = "none"
                if loop < (len(vector_list)-1):
                    next_vector_name = vector_names_list[loop+1]
                    #pdb.gimp_message(next_vector_name)
                    if (len(next_vector_name) == 6 and " part" in next_vector_name): # look to see if the next vector needs joining onto the it
                        part_vector_id = vector_list[loop + 1]
                        # get paths etc associated with the part vector
                        num_part_strokes, part_vector_list = pdb.gimp_vectors_get_strokes(part_vector_id)
                        stroke_loop = 0
                            
                        while stroke_loop<num_part_strokes:
                            # loop through paths that make up the part vector
                            part_vector_stroke = part_vector_list[stroke_loop]
                            # get information associated with the current vector part
                            TypeOfStrokeInPart_Vector,NumberOfPointsInPart_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(part_vector_id, part_vector_stroke)
                            part_ListOfControlPoints = ListOfControlPoints
                            # add current part path details to the base character
                            vector_id = pdb.gimp_vectors_stroke_new_from_points(vector_list[loop], 0, NumberOfPointsInPart_Vector, ListOfControlPoints, IsStrokeClosed)
							
                            stroke_loop += 1
				
				
                        pdb.gimp_image_remove_vectors(image,  vector_list[loop+1])
            loop += 1
		
        ###########
		
        old_handler = pdb.gimp_message_get_handler()		
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message("This file contained some two part character paths: They have been converted to single paths to allow dynamics emulation to be applied.")
        pdb.gimp_message_set_handler(old_handler)	
        # update list of vectors to accomodate any removed vectors
        num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
        vector_list = image.vectors
			
    layer_list = image.layers
    
    # create a list of the actual layer names
    loop = 0
    while loop<len(layer_list):
        pdb.gimp_item_set_visible(layer_list[loop], False)
        layer_name = pdb.gimp_item_get_name(layer_list[loop])
        layer_name = layer_name.decode('utf-8')
        if loop == 0:
            layer_names_list = [layer_name]
        else:
            layer_names_list.append(layer_name)
        loop += 1
    found_valid_vector = False
    found_vector = False
    if num_vectors>0: # if there are any vectors in this image do something
        found_vector = True
        loop = 0
        update_canvas_size = False
        previous_name = "undefined"
			
        while loop<num_vectors:
            vector_id = vector_list[loop]
            active_vector = pdb.gimp_image_set_active_vectors(image,vector_id)
            active_vector_name = pdb.gimp_item_get_name(vector_id)
            active_vector_name = active_vector_name.decode('utf-8')
				
            ###############
				
            if (len(active_vector_name) == 6 and " char" in active_vector_name) or ((len(active_vector_name) == 6 and " part" in active_vector_name) and previous_name == active_vector_name[0]): # check for compatible paths
			
			
                first_char = active_vector_name[0]
                #pdb.gimp_message("first_char = "+str(first_char))
				
                if first_char in thinner_chars:
                    pdb.gimp_context_set_line_width(thinner_line_size)
                    pdb.gimp_context_set_brush_size(thinner_brush_size)
				
                else:
                    pdb.gimp_context_set_line_width(line_width_major)
                    pdb.gimp_context_set_brush_size(brush_width_major)
					
                found_valid_vector = True
				
                if loop<(num_vectors-1):
                    next_vector_id = vector_list[loop + 1]
                    next_vector_name = pdb.gimp_item_get_name(next_vector_id)
                    next_vector_name = next_vector_name.decode('utf-8')
                else:
                    next_vector_id = -1
                    next_vector_name = "undefined"
				
                if " char" in active_vector_name:
                    new_name = active_vector_name[0]
                    previous_name = new_name
                    new_name = new_name.decode('utf-8')
                    new_name += ".png"
					
                    
                    layer_exists = False
                    if new_name in layer_names_list: #check to see if there is already a layer of this name
                        # and get its layer id
                        existing_layer = pdb.gimp_image_get_layer_by_name(image, new_name)
                        layer_exists = True
                    
                    # use image height for layer and get information about the width from points along path - hopefully they will help define the width fairly accurately
                    NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(vector_id)
                    maxX,minX,stroke_loop = -3000000,3000000,0
                    while stroke_loop<NumberOfStrokes_in_Path:
                        Char_Vector_Stroke = Char_Vector_StrokeIDs_List[stroke_loop]
                        TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(vector_id,Char_Vector_Stroke)
                        ploop,cloop = 0,0
                        #make a copy of the returned list of control points
                        Temp_ListOfControlPoints = list(ListOfControlPoints)
                        # remove unnecessary control handle entries retaining x coordinate positions for points
                        while ploop<len(ListOfControlPoints):
                            if ploop == 2 or ((ploop-2)%6 == 0):
                                cloop -= 0
                            else:
                                list_pos = ploop - cloop
                                del Temp_ListOfControlPoints[list_pos]
                                cloop += 1
                            ploop+=1
							
                        # get maximum x position from this list to find layer width
                        if max(Temp_ListOfControlPoints) > maxX: maxX = max(Temp_ListOfControlPoints)
                        # get minimum x position to use to help set layer x offset
                        if min(Temp_ListOfControlPoints) < minX: minX = min(Temp_ListOfControlPoints)
                        stroke_loop += 1
						
                    # use line width or brush width to get a path width according to type of current stroke method
                    if stroke_method == 1:
                        stroke_width = int(brush_width_major/2)
                    else:
                        stroke_width = int(line_width_major/2)
							
                    # set the width of the layer                        
                    layer_left_margin = minX - ((stroke_width) + left_margin) # define x pos of left edge of layer
                    layer_right_margin = maxX + ((stroke_width) + right_margin) # define x pos of right edge of layer
                    y_offset, char_x_offset = 0, 0

                    if layer_left_margin<0:
                        layer_width = abs(layer_left_margin - layer_right_margin)
                        char_x_offset = (layer_left_margin)
                    else:
                        layer_width = layer_right_margin - layer_left_margin
                        char_x_offset = (layer_left_margin)
							
							
                    new_layer = pdb.gimp_layer_new(image, layer_width, image.height, 1, new_name,100,0)
                    pdb.gimp_image_insert_layer(image, new_layer, None, 0)
                    if layer_width>new_layer.width:
                        pdb.gimp_layer_resize(new_layer, layer_width, image.height, 0, 0)
                        if layer_width>image.width: update_canvas_size = True
							
						
                    # move layer to stroke position
                    pdb.gimp_item_transform_translate(new_layer, char_x_offset, y_offset)


                elif " part" in active_vector_name:
                    # check to see if this is a part layer to stroke in the secondary size
                    new_name = active_vector_name[0] + ".png"
                    part_name = active_vector_name[0]
                    new_name = new_name.decode('utf-8')
                    if not layer_exists: new_layer = pdb.gimp_image_get_layer_by_name(image, new_name)
                    if " part" in active_vector_name:
                        # if the path is one of ours set brush/line size
                        pdb.gimp_context_set_line_width(line_width_minor)
                        pdb.gimp_context_set_brush_size(brush_width_minor)
						
                        # use image height for layer and get information about the width from points along path - hopefully they will help define the width fairly accurately
                        part_maxX,part_minX,stroke_loop = -3000000,3000000,0
                        NumberOfStrokes_in_Path,Char_Vector_StrokeIDs_List = pdb.gimp_vectors_get_strokes(vector_id)
                        while stroke_loop<NumberOfStrokes_in_Path:
                            Char_Vector_Stroke = Char_Vector_StrokeIDs_List[stroke_loop]
                            TypeOfStrokeInChar_Vector,NumberOfPointsInChar_Vector,ListOfControlPoints,IsStrokeClosed = pdb.gimp_vectors_stroke_get_points(vector_id,Char_Vector_Stroke)
                            ploop,cloop = 0,0
                            #make a copy of the returned list of control points
                            Temp_ListOfControlPoints = list(ListOfControlPoints)
                            # remove unnecessary control handle entries and retain x coordinate positions for points
                            while ploop<len(ListOfControlPoints):
                                if ploop == 2 or ((ploop-2)%6 == 0):
                                    cloop -= 0
                                else:
                                    list_pos = ploop - cloop
                                    del Temp_ListOfControlPoints[list_pos]
                                    cloop += 1
                                ploop+=1
                            # get maximum x position to use to help set layer width
                            if max(Temp_ListOfControlPoints) > part_maxX: part_maxX = max(Temp_ListOfControlPoints)
                            # get minimum x position to use to help set layer x offset
                            if min(Temp_ListOfControlPoints) < part_minX: part_minX = min(Temp_ListOfControlPoints)
                            stroke_loop += 1
							
                        # use line width or brush width to stroke a path
                        if stroke_method == 1:
                            stroke_width = int(brush_width_minor/2)
                        else:
                            stroke_width = int(line_width_minor/2)

                        # check if the part layer will need a wider layer than the parent character layer
                        # set the margin size for the part layers
                        part_layer_right_margin = part_maxX + ((stroke_width) + right_margin) # define x pos of right edge of part layer
                        part_layer_left_margin = part_minX - ((stroke_width) + left_margin) # define x pos of left edge of part layer
                        part_width = layer_width # provisional width
                        
                        if min(layer_left_margin, part_layer_left_margin)<0:
                            part_width = abs(min(layer_left_margin, part_layer_left_margin) - max(layer_right_margin,part_layer_right_margin) )
                        else:
                            part_width = max(layer_right_margin,part_layer_right_margin) - min(layer_left_margin, part_layer_left_margin)
                        # set value for this part layer offset
                        part_x_offset = 0
                        if part_layer_left_margin<layer_left_margin:
                            part_x_offset = abs(part_layer_left_margin - layer_left_margin)
                        else:
                            part_x_offset = 0
							
                        if part_width>layer_width:
                            # resize the character layer if extra width is required
                            pdb.gimp_layer_resize(new_layer, part_width, image.height,part_x_offset, y_offset)
                        pdb.gimp_item_transform_translate(new_layer, char_x_offset , y_offset) # move the layer into position for stroking
						
                        if part_width>image.width: update_canvas_size = True # set flag for updating canvas size
											
                # make layer visible
                pdb.gimp_item_set_visible(new_layer, True)
		
				
					
                # stroke the path (using current width/colour settings) the required number of times
                repeat_loop = 0
					
                #if force_dynamics and stroke_method == 1 and " char" in active_vector_name: # move current vector to accomodate changes in left hand margins for track direction option
                if force_dynamics and stroke_method == 1: # move current vector to accomodate changes in left hand margins for track direction option

                    num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(vector_id)
                
                    for i in range(0,num_strokes):
                        pdb.gimp_vectors_stroke_translate(vector_id, stroke_ids[i], (0-layer_left_margin), 0)
                    
                    '''
                    if force_dynamics and stroke_method == 1 and " part" in active_vector_name: # move current vector to accomodate changes in left hand margins for track direction option
                    #if force_dynamics and stroke_method == 1: # move current vector to accomodate changes in left hand margins for track direction option

                        num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(vector_id)
                
                        for i in range(0,num_strokes):
                            #pdb.gimp_vectors_stroke_translate(vector_id, stroke_ids[i], (0-layer_left_margin), 0)
                            pdb.gimp_vectors_stroke_translate(vector_id, stroke_ids[i], (0+(int(brush_width_minor)/2)), 0)
                    '''
                
                
                while repeat_loop<repeat_process:
                    #pdb.gimp_image_undo_group_start(image)
				
				
                    if (force_dynamics and stroke_method == 1):
					
                        # Using a version of the code that Tim made available for using the paintbrush tool to follow direction of a stroke which in turn uses a function written by Ofnuts
                        #pdb.gimp_context_set_dynamics("Track Direction") # Doesn't seem to take effect at all when we try to stroke vectors
                        
                        num_strokes, stroke_ids = pdb.gimp_vectors_get_strokes(vector_id)
                        paint_between_length = brush_spacing*50
                        original_brush_size = pdb.gimp_context_get_brush_size()
                        target_brush_size = float(original_brush_size-((original_brush_size/100)*min_track_size ))
                        target_brush_size = float(((original_brush_size/100)*min_track_size ))
                        for i in range(0,num_strokes):
                            step_brush_size = original_brush_size
                            stroke_length = pdb.gimp_vectors_stroke_get_length(vector_id,stroke_ids[i],0.01)
                            paint_between_length = float(brush_spacing*50)
                            num_brush_steps = float(stroke_length/paint_between_length)
                            if stroke_length % brush_spacing == 0: paint_between_length -= 1
                            dist_at = 0.0
                            if stroke_length > 0.0:size_step = float((original_brush_size-target_brush_size)/num_brush_steps)
                            
                            while dist_at < (stroke_length-1):
                                x1,y1,slope,valid = pdb.gimp_vectors_stroke_get_point_at_dist(vector_id,stroke_ids[i],dist_at-1,0.01)
                                x2,y2,slope,valid = pdb.gimp_vectors_stroke_get_point_at_dist(vector_id,stroke_ids[i],dist_at,0.01)
                                x3,y3,slope,valid = pdb.gimp_vectors_stroke_get_point_at_dist(vector_id,stroke_ids[i],dist_at+1,0.01)
         
                                points = []
                                points.append(x2)
                                points.append(y2)
                                dist_at += paint_between_length
								
                                if track_direction:
                                    angle = computeOrientedSlope(x3-x1,y3-y1,slope) #calculate angle based on slope using function by Ofnuts
                                    angle = angle/2/math.pi * 360
                                    if angle > 180:
                                        angle -= 360
                                    pdb.gimp_context_set_brush_angle(angle)
									
								
                                if random_angle:
                                    random.seed()
                                    angle = (random.randrange(0, 360)-180.0)
                                    pdb.gimp_context_set_brush_angle(angle)
								
									
                                if size_jitter>0:
                                    if stroke_width>10:
                                        jitter_stroke_width = stroke_width * 2
                                        jitter_adj = int((float(jitter_stroke_width)/100)* size_jitter) 
                                        jitter_size_min, jitter_size_max = jitter_stroke_width - jitter_adj, jitter_stroke_width + jitter_adj
                                        random.seed()
                                        jitter_brush_size = random.randrange(jitter_size_min, jitter_size_max)
                                        if jitter_brush_size<10:jitter_brush_size = 10
                                        pdb.gimp_context_set_brush_size(jitter_brush_size)
										
										
                                if space_jitter>0:
                                    jitter_space_adj = ((brush_spacing*50)/100) * space_jitter
                                    jitter_space_min, jitter_space_max = (brush_spacing*50) - jitter_space_adj, (brush_spacing*50) + jitter_space_adj
                                    random.seed()
                                    jitter_brush_spacing = random.randrange(int(jitter_space_min), int(jitter_space_max))
                                    if jitter_brush_spacing<1:jitter_brush_spacing = 1
                                    paint_between_length = jitter_brush_spacing
									
                                if first_char not in exclude_chars: # is the character in the exclusion string
                                    if ((track_size and i>0) or (track_size and (i==0 and track_size_all))): # check to see for taper option
                                        step_brush_size -= size_step
                                        if step_brush_size<1.00: step_brush_size = 1.0
                                        pdb.gimp_context_set_brush_size(step_brush_size)

									
                                pdb.gimp_paintbrush(new_layer,0,len(points),points,PAINT_CONSTANT,0) # use paintbrush tool to paint along current stroke to emulate track direction
							
					
                    else:
                        pdb.gimp_drawable_edit_stroke_item(new_layer, vector_id) # stroke path when no track direction is used
					
                    #pdb.gimp_image_undo_group_end(image)
                    pdb.gimp_displays_flush()
                    repeat_loop += 1
					
					
                # reset the layer position
                new_layer.set_offsets(0, 0)
				
				
                if "part" not in next_vector_name:
                #if "part" not in next_vector_name and layer_exists:
                    pdb.gimp_item_set_visible(new_layer, True)
                    if layer_exists:
                        pdb.gimp_item_set_visible(existing_layer, True)
						
                    if add_drop:
                        drop_layer = new_layer
                        new_layer = proc_dropshadow(image, drop_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour)
						
                    if layer_exists: pdb.gimp_image_merge_down(image, new_layer, 0)
						
							
                    new_layer = image.active_layer 
						
				
                # turn off layer visibility
                if loop <> (num_vectors-1): pdb.gimp_item_set_visible(new_layer, False) # turn off layer visibility
                pdb.gimp_displays_flush()
                '''		
                '''
            loop +=1
				
		
        if update_canvas_size: pdb.gimp_image_resize_to_layers(image)
		
        # check to see if there is an existing space layer and create one if necessary
        if not "space.png"  in layer_names_list and found_valid_vector == True:
            space_width = (float(image.width) / 100) * space_width
            space_layer = pdb.gimp_layer_new(image,(space_width), image.height, 1, "space.png",100,0)
            pdb.gimp_image_insert_layer(image, space_layer, None, 0)
            # make space layer not visible
            pdb.gimp_item_set_visible(space_layer, False) # turn off layer visibility
        if found_valid_vector: pdb.gimp_image_set_active_layer(image, new_layer)
        if found_valid_vector: pdb.gimp_item_set_visible(new_layer, True) # turn off layer visibility

		
			
        if found_valid_vector == False: # Report no suitable paths
            old_handler = pdb.gimp_message_get_handler()		
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message("This image has at least one path but there are none suitable for the plug-in to work with.")	
            pdb.gimp_message_set_handler(old_handler)
        else:
            if remove_paths:
                # get lists of vector ids in this image
                num_vectors, vector_ids = pdb.gimp_image_get_vectors(image)
                vector_list = image.vectors
                loop = 0
                while loop<num_vectors:
                    vector_id = vector_list[loop]
                    active_vector_name = pdb.gimp_item_get_name(vector_id)
                    active_vector_name = active_vector_name.decode('utf-8')
                    if (len(active_vector_name) == 6 and " char" in active_vector_name) or ((len(active_vector_name) == 6 and " part" in active_vector_name)): 
                        pdb.gimp_image_remove_vectors(image, vector_list[loop])
                    loop += 1
           
	
        if found_vector == False: # Report no paths
            old_handler = pdb.gimp_message_get_handler()		
            pdb.gimp_message_set_handler(MESSAGE_BOX)
            pdb.gimp_message("There are no paths, associated with this image.")	
            pdb.gimp_message_set_handler(old_handler)	
			
			
    
	
######################################################################################################################################
   
#Function by ofnuts' from ofn-text-along-path.py
def computeOrientedSlope(dx,dy,slope): #This is a better version than below computeThetaWithSlope function.
    if abs(slope) > 100000: # very vertical
        return math.atan2(dy,dx) # no perfect, but properly oriented
        # keep dx/dy signs but give them the same ratio as in slope
    return math.atan2(math.copysign(slope,dy),math.copysign(1,dx))

######################################################################################################################################

def proc_dropshadow(im, target_layer, drop_offset_x, drop_offset_y, drop_blur, drop_opacity, drop_colour):
    #REM Create a text drop shadow effect - optionally clipped to the text layer for transparent type texts
	
    # make a copy of the image width/height
    old_im_width, old_im_height = im.width, im.height
    name = pdb.gimp_item_get_name(target_layer)
    # Copy layer to use as the drop shadow
    shadow_layer = pdb.gimp_layer_copy(target_layer, False)
    # Add and re-position the layer in the layer stack
    im.add_layer(shadow_layer, -1)
    pdb.gimp_image_lower_item(im, shadow_layer)
    # get layer offsets for the shadow layer (which will also be the same as the original layer holding the user text)
    shadow_offset_x, shadow_offset_y = pdb.gimp_drawable_offsets(shadow_layer)
    # work out the centre point of the shadow/original layer
    shad_centre_x = (shadow_layer.width/2)
    shad_centre_y = (shadow_layer.height/2)
    # make up new size for the shadow layer to accommodate blurred result - currently twice the existing width and height
    new_shad_width = shadow_layer.width*2
    new_shad_height = shadow_layer.height*2
    # work out new offsets for the new size of the shadow layer to ensure that the shadow and original layers are still aligned
    new_shad_layer_offset_x = 0 - ((shadow_layer.width - new_shad_width)/2)
    new_shad_layer_offset_y = 0 - ((shadow_layer.height - new_shad_height)/2)
    # and apply the new offsets to the shadow layer
    pdb.gimp_layer_resize(shadow_layer, new_shad_width, new_shad_height, new_shad_layer_offset_x, new_shad_layer_offset_y)
    # make a note of the new shadow layer offsets
    shadow_resized_offset_x, shadow_resized_offset_y = pdb.gimp_drawable_offsets(shadow_layer)	
	
    # if any of the shadow layer falls outside the image boundaries
    reset_shadow = False
    if new_shad_layer_offset_x<0 or new_shad_layer_offset_y<0 or shadow_layer.width > im.width or shadow_layer.height > im.height:
        reset_shadow = True
        # set offsets of the shadow layer to image origin
        pdb.gimp_layer_set_offsets(shadow_layer,0,0)
        # set offsets for original text layer to match
        pdb.gimp_layer_set_offsets(target_layer,0 + abs(new_shad_layer_offset_x) ,0 + abs(new_shad_layer_offset_y))
        # resize the image to accomodate the new shadow layer size
        new_im_width, new_im_height = im.width, im.height
        if shadow_layer.width > im.width: new_im_width = shadow_layer.width
        if shadow_layer.height > im.height: new_im_height = shadow_layer.height
        pdb.gimp_image_resize(im, new_im_width, new_im_height, 0, 0)
		
    # get alpha selection
    pdb.gimp_image_select_item(im, 2, shadow_layer)
    # fill the text on the shadow layer with the background colour
    if drop_colour<2:
        pdb.gimp_drawable_edit_fill(shadow_layer,drop_colour)
    # deselect
    pdb.gimp_selection_none(im)
    # blur shadow 
    pdb.plug_in_gauss(im, shadow_layer, drop_blur, drop_blur, 1)
    # crop the blurred layer
    pdb.plug_in_autocrop_layer(im, shadow_layer)
    # move the shadow layer to user position and adjust opacity
    temp = pdb.gimp_layer_translate(shadow_layer, drop_offset_x, drop_offset_y)
    pdb.gimp_layer_set_opacity(shadow_layer, drop_opacity)
    # merge and rename
    target_layer = pdb.gimp_image_merge_down(im,target_layer,0)
    pdb.gimp_item_set_name(target_layer, name)
    if reset_shadow: 
        cropped_offset_x, cropped_offset_y = pdb.gimp_drawable_offsets(target_layer)	
        pdb.gimp_layer_set_offsets(target_layer,shadow_offset_x, shadow_offset_y)
    pdb.gimp_image_resize(im, old_im_width, old_im_height, 0, 0)
    pdb.gimp_image_resize_to_layers(im)
    return target_layer
	
######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################
	
register(
    "python_fu_Paths_To_Layers_v0-09_T",
    "Provides options for turning paths into character layers ",
    "Provides options for turning paths into character layers ",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font Paths to Layers v.0.09 Tabbed...",
    "RGBA",
    [
    # 1st Tab stroke options for line and brush
    (PF_OPTION, "option", "Stroke path with: ", 0, ("Brush", "Line")),
    (PF_SPINNER, "brush_primary", "Primary stroke size:", 70, (6, 2000, 2)),
    (PF_SPINNER, "brush_secondary", "Secondary stroke size (as %):", 76, (10, 100, 2)),
    (PF_SPINNER, "brush_spacing", "Brush spacing:", 50, (0, 5000, 1)),
    (PF_STRING, "thinner_string", "Characters to stroke\nin the secondary width:", "><|}{+_)(*&^%$#@!~/\][=-:.,';`"+str( unichr(8364))+str( unichr(162))+str( unichr(163))+str( unichr(169))+str( unichr(174))+str( unichr(188))+str( unichr(189))+str( unichr(190))+str( unichr(215))+str( unichr(247))+str( unichr(34))+str( unichr(176))+str( unichr(178))+str( unichr(179))+str( unichr(183))+str( unichr(185))),
    (PF_OPTION, "cap_option", "Path end cap style: ", 1, ("Butt", "Round", "Square")),
    (PF_OPTION, "join_option", "Path line join style: ", 1, ("Mitre", "Round", "Bevel")),
    (PF_SPINNER, "left_margin", "Left hand character margin adjustment:", 10, (-500, 500, 2)),
    (PF_SPINNER, "right_margin", "Right hand character margin adjustment:", 10, (-500, 500, 2)),
    (PF_SPINNER, "space_size", "Width of space character\n(as % of image width):", 50, (10, 100, 1)),
    (PF_SPINNER, "repeats", "Repeat strokes:", 3, (1, 50, 1)),
	
	
    # 2nd Tab brush specific options
    (PF_TOGGLE, "dynamics_option", "Emulate brush dynamics: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "dynamics_track", "Track direction: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "brush_angle", "Random brush angle: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_SPINNER, "jitter_adj_amount", "Brush size jitter (%):", 0, (0, 50, 1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "jitter_spacing_amount", "Brush spacing jitter (%):", 0, (0, 50, 1),{'newLine':'1','tab':1}),
    (PF_TOGGLE, "size_track", "Track size (taper strokes): ", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_SPINNER, "track_size_min", "Minimum taper size (as %):", 50, (1, 100, 5),{'newLine':'1','tab':1}),
    (PF_TOGGLE, "size_track_all", "Taper all strokes in characters: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_STRING, "exclude_string", "Do not use taper with these characters: ", ("OoQ08%"+str( unichr(169))+str( unichr(174))), 0,{'newLine':'1','tab':1}),
	
    # 3rd Tab drop shadow options
    (PF_SPINNER, "pf_drop_offset_x", "Shadow offset x position: ", 0, (-50, 50, 1),{'newLine':'1','tab':2}),
    (PF_SPINNER, "pf_drop_offset_y", "Shadow offset y position: ", 0, (-50, 50, 1),{'newLine':'1','tab':2}),
    (PF_SPINNER, "pf_drop_blur", "Shadow blur percentage: ", 15, (0, 50, 1),{'newLine':'1','tab':2}),
    (PF_SPINNER, "pf_drop_opacity", "Shadow opacity: ", 70, (20, 100, 1),{'newLine':'1','tab':2}),
    (PF_OPTION, "drop_colour", "Drop shadow", 1, ( "Foreground Colour", "Background colour", "Layer copy"),{'newLine':'1','tab':2}),
	
    # 4rd tab path and preview options
    (PF_TOGGLE, "remember", "Remember initial stroke width: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "reset_option", "Reset initial stroke width\n(Removes all character layers): ", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "path_option", "Remove character paths only \nfrom custom font file: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_TOGGLE, "Preview_option", "Preview single character\n(to a new image): ", 0, [("No"),("Yes") ],{'newLine':'1','tab':2}),
    (PF_STRING, "preview_string", "Preview character: ", ("A"), 0,{'newLine':'1','tab':2}),
	
    (PF_TAB    , "ptab"    , "tabs", 0, (_('Stroking options'),_('Brush options'),_('Drop shadow & Paths')) ),
    ],
    [],
    do_stuff)

main()
